//-------------------------------------------------------------------
//	File:		   mesh.cpp
//	Comments:   by George Ziniewicz
//-------------------------------------------------------------------

#include "mesh.h"
#include "bump.h"
#include "physics.h"
#include "hardware.h"

//-------------------------------------------------------------------

Mesh::Mesh() {
	nodesx=1;
	nodesy=1;
	sizex=1;
	sizey=1;
	texmagx=1;
	texmagy=1;
	texoffx=0;
	texoffy=0;
	meshedit=0; meshx=0; meshy=0;
	shapefreq=1;
	shapeclipy1=0;
	height=1;
	min=0;
	max=1;
	useDisplayList=0;
	listID=0;
	r=g=b=a=1;
	colors=0;
	normals=0;
	cullStrip=0;
	return;
};
Mesh::~Mesh() {
};

//--------------------------------------------------------------------

void Mesh::newmesh(int innodesx, int innodesy, float insizex, float insizey) {
	nodesx=innodesx;
	nodesy=innodesy;
	sizex=insizex;
	sizey=insizey;
	init();
}

//--------------------------------------------------------------------

void Mesh::init(void) {
	int ix,iy;
	float sizexh,sizeyh, fx,fy, cellsizex,cellsizey;

	sizexh=sizex*.5;
	sizeyh=sizey*.5;
	cellsizex=sizex/nodesx;
	cellsizey=sizey/nodesy;
	for    (ix=0; ix<=nodesx; ix++) {
		for (iy=0; iy<=nodesy; iy++) {
			fx=ix*cellsizex-sizexh;
			fy=iy*cellsizey-sizeyh;
			h[ix][iy][0]= fx;
			h[ix][iy][1]=0;
			h[ix][iy][2]=-fy;
			n[ix][iy][0]=0;
			n[ix][iy][1]=1;
			n[ix][iy][2]=0;
			//c[ix][iy][0]=1;
			//c[ix][iy][1]=1;
			//c[ix][iy][2]=1;
			//c[ix][iy][3]=1;
			i[ix][iy]=1;
		}
		d[ix][0]=d[ix][1]=-1;
	}
	height=0;
	maxheight=0;
	min=0;
	max=0;
	texgen();
	if (listID>0) {
		glDeleteLists(listID,1);
		listID=0;
	}
	//calc_normal();
	//calc_shadow();
}

//----------------------------------------------------------------------------

void Mesh::init(Bump *bump, float scale, float offset) {
	int ix,iy;
	float sizexh,sizeyh, fx,fy, tx,ty, cellsizex,cellsizey, ht;

	sizexh=sizex*.5;
	sizeyh=sizey*.5;
	cellsizex=sizex/nodesx;
	cellsizey=sizey/nodesy;
	min=9999;
	max=-9999;
	maxheight=-9999;
	for	 (ix=0; ix<=nodesx; ix++) {tx=(float)ix/nodesx; fx=ix*cellsizex-sizexh;
		for (iy=0; iy<=nodesy; iy++) {ty=(float)iy/nodesy; fy=iy*cellsizey-sizeyh;
			ht=bump->get(tx,1-ty)*scale+offset;
			min=minf(min,ht);
			max=maxf(max,ht);
			maxheight=maxf(maxheight,ht);
			h[ix][iy][0]= fx;
			h[ix][iy][1]= ht;
			h[ix][iy][2]=-fy;
			n[ix][iy][0]=0;
			n[ix][iy][1]=1;
			n[ix][iy][2]=0;
			//c[ix][iy][0]=1;
			//c[ix][iy][1]=1;
			//c[ix][iy][2]=1;
			//c[ix][iy][3]=1;
			i[ix][iy]=1;
		}
		//d[ix][0]=d[ix][1]=-1;
	}
	height=max-min;
	//calc_normal();
	texgen();
	if (listID>0) {
		glDeleteLists(listID,1);
		listID=0;
	}
	//calc_shadow();
}

//------------------------------------------------------------------

void Mesh::close(void) {
	char s[80];
	if (listID>0) {
		sprintf(s,"   Destroying mesh display list # %u...",listID);
		log_message(s);
		glDeleteLists(listID,1);
		listID=0;
	}
}

//------------------------------------------------------------------

void Mesh::draw(float size) {
	int ix, nodesyh=nodesy*.5;
	Hw *hw=Hw::get();
	float cc[4],r,g,b;

	glGetFloatv(GL_CURRENT_COLOR,cc);
	r=cc[0];
	g=cc[1];
	b=cc[2];

	if (useDisplayList) {
		if (listID==0) {         // not made?
			listID=glGenLists(1);
			if (listID==0) return;
			glNewList(listID,GL_COMPILE_AND_EXECUTE);
		} else {                 // been made?
			glCallList(listID);
			Hw::get()->pref.ntris+=nodesx*nodesy;
			return;
		}
	}
	if (normals==0) glNormal3f(0,1,0); 
	for (ix=0; ix< nodesx; ix++) { // do all y strips across in x
		if (cullStrip==0) do_strip(ix,0,nodesy,r,g,b);
		else {
			for (int i=0; i<8; i++) { // cull strip sections
				if (cull_strip(ix,i/8.*nodesy, (i+1)/8.*nodesy)==0)
					do_strip(ix,i/8.*nodesy,(i+1)/8.*nodesy,r,g,b);
			}
		}
	}
	if (useDisplayList) glEndList();
}

//----------------------------------------------------------------------------

float Mesh::get(float x, float y) { // 0-1
	int xi, yi;
	float v, v0, xrem,yrem;

	if      (x<0)  x=0;
	else if (x>.999) x=.999;
	if      (y<0)  y=0;
	else if (y>.999) y=.999;
	//x=wrap(x,0,1);
	//y=wrap(y,0,1);
	//if (x>=1) x=.999;
	//if (y>=1) y=.999;

	xi=x*nodesx; 
	yi=y*nodesy;
	xrem=x*nodesx-xi;
	yrem=y*nodesy-yi;
	if ((xrem+yrem)<1) {
		v=v0=h[     xi          ][     yi          ][1];
		v+= (h[mini(xi+1,nodesx)][     yi          ][1]-v0)*xrem;
		v+= (h[     xi          ][mini(yi+1,nodesy)][1]-v0)*yrem;
	} else {
		v=v0=h[mini(xi+1,nodesx)][mini(yi+1,nodesy)][1];
		v+= (h[     xi          ][mini(yi+1,nodesy)][1]-v0)*(1-xrem);
		v+= (h[mini(xi+1,nodesx)][     yi          ][1]-v0)*(1-yrem);
	}
	return v;
}

//--------------------------------------------------------------------

void Mesh::getv(float x, float y, float *xi, float *zi) {
	float vx,vy,vz, len;

	vx=(get(x+.0001,y     )-get(x-.0001,y     ))*height; ////////////////////
	vy=2*sizex;
	vz=(get(x,     y+.0001)-get(x,     y-.0001))*height;
	len=sqrt(vx*vx+vy*vy+vz*vz);
	*xi=-vx/len;
	*zi=-vz/len;
}

//--------------------------------------------------------------------

void Mesh::shape(int type, float scale, float offset, float freq) {
	float x,y, mag, dist,s2=nodesx*.5, ht;
	int ix, iy, nodesh=nodesx*.5;
	shapefreq=freq;
	min=9999;
	max=-9999;
	maxheight=-9999;

	switch(type) {
		case SINBUMPS: mag=(3*pi2)/sizex; break;
		case PARABOLA: mag=  1./nodesx;   break;
		case CENTERSQ: mag=(20./nodesx);  break;
		case RINGS:    mag=(3*pi2)/sizex; break;
		case COSBUMP:  mag=(3*pi2)/sizex; break;
	}
	for    (ix=0; ix<=nodesx; ix++) {
		for (iy=0; iy<=nodesy; iy++) {
			switch(type) {
				case SINBUMPS: // sin bumps
					h[ix][iy][1]+=(sinf(h[ix][iy][0]*mag*shapefreq)-sinf(h[ix][iy][2]*mag*shapefreq))*scale+offset;
					break;
				case PARABOLA: // spherical bend
				   h[ix][iy][1]+=( (ix-nodesh)*(ix-nodesh)+
					                (iy-nodesh)*(iy-nodesh) )*mag*scale+offset;
					break;
				case CENTERSQ:  // center sq
					if ((fabs(ix-nodesx*.5)<2)&&
					    (fabs(iy-nodesy*.5)<2)) h[ix][iy][1]+=mag*scale+offset;
				   else                        h[ix][iy][1]+=0;
					break;
				case RINGS:
					x=h[ix][iy][0];
					y=h[ix][iy][2];
					dist=sqrt(x*x+y*y);
					ht=cosf(dist*mag*shapefreq)*scale+offset;
					if (ht>shapeclipy1) ht=shapeclipy1;
					h[ix][iy][1]+=ht;
					break;
				case COSBUMP:
					h[ix][iy][1]=offset;
					break;
			}
			min=minf(min,h[ix][iy][1]);
			max=maxf(max,h[ix][iy][1]);
			maxheight=maxf(maxheight,h[ix][iy][1]);
		}
	}
	height=max-min;
	if (type==COSBUMP) depress(.5*nodesx,.5*nodesy, .5*nodesx*shapefreq, scale);
}

//------------------------------------------------------------------

void Mesh::spherical_map(float scale) {
	int ix, iy;
	float rx=pi2/nodesx, ax,ay;
	float ry=pi1/nodesy;

	for    (iy=0; iy<=nodesy; iy++) {
		ay=iy*ry;
		for (ix=0; ix<=nodesx; ix++) {
			ax=ix*rx;
			h[ix][iy][0]= sinf(ax)*sinf(ay)*scale;
			h[ix][iy][1]=         -cosf(ay)*scale;
			h[ix][iy][2]= cosf(ax)*sinf(ay)*scale;
			n[ix][iy][0]= sinf(ax)*sinf(ay);
			n[ix][iy][1]=         -cosf(ay);
			n[ix][iy][2]= cosf(ax)*sinf(ay);
		}
	}
	min=1; max=2; height=1;
}

//------------------------------------------------------------------

void Mesh::texgen(void) {
	float tx, tmagx=(texmagx/nodesx),
		       tmagy=(texmagy/nodesy);

	for    (int ix=0; ix<=nodesx; ix++) {
		tx             =ix*tmagx+texoffx;
		for (int iy=0; iy<=nodesy; iy++) {
			t[ix][iy][0]=tx;
			t[ix][iy][1]=iy*tmagy+texoffy;
			t[ix][iy][2]=0;
			t[ix][iy][3]=1;
		}
	}
}

//------------------------------------------------------------------

void Mesh::bumpgen(void) {
	int nodesxh=nodesx*.5,nodesyh=nodesy*.5,dx,dy;
	float tx, tmagx=(texmagx/nodesx),
		       tmagy=(texmagy/nodesy), v,nx,ny,len;//,d;
	Hw *hw=Hw::get();

	for    (int ix=0; ix<=nodesx; ix++) {
		tx             =ix*tmagx+texoffx;
		dx=ix-nodesxh;
		for (int iy=0; iy<=nodesy; iy++) {
			dy=iy-nodesyh;
			len=sqrt(dx*dx+dy*dy);
			if (len) {
				nx=dx/len;
				ny=dy/len;
			} else {
				nx=0; ny=0;
			}
			v=(sinf(ix*hw->pref.paren)+
				sinf(iy*hw->pref.paren))*.03*hw->pref.spd1+hw->pref.spd2;

			//d=sinf(len*hw->pref.paren+hw->time.spd1);
			//v=d*hw->pref.spd2;
			t[ix][iy][3]=v;
		}
	}
}

//------------------------------------------------------------------

void Mesh::depress(int inx, int iny, int insize, float indepth) {
	int x,y, xa,ya;
	float dist, s2, ht;
	float m[100][100];

	s2=insize*.5;
	for    (x=0;x<nodesx;x++) {
		for (y=0;y<nodesy;y++) {
			m[x][y]=h[x][y][1];
		}
	}
	for    (x=-s2; x<s2; x++) {
		for (y=-s2; y<s2; y++) {
			xa=(x+inx)%nodesx;
			ya=(y+iny)%nodesy;
			dist=sqrt(x*x+y*y);
			if (dist>=s2) continue;
			ht=cos_interp(indepth,0,dist/s2);
			switch(sign(indepth)){
			case -1:
			if ((h[xa][ya][1]+ht)<m[xa][ya])
				  m[xa][ya]=       h[xa][ya][1]+ht;
			break;
			case 1:
			if ((h[xa][ya][1]+ht)>m[xa][ya])
				  m[xa][ya]=       h[xa][ya][1]+ht;
				break;
			}
		}
	}
	min=9999;
	max=-9999;
	maxheight=-9999;
	for    (x=0;x<nodesx;x++) {
		for (y=0;y<nodesy;y++) {
			h[x][y][1]=m[x][y];
			min=minf(min,h[x][y][1]);
			max=maxf(max,h[x][y][1]);
			maxheight=maxf(maxheight,h[x][y][1]);
		}
	}
	height=max-min;
}

//---------------------------------------------------------------------

int Mesh::inMesh(float x,float y) {
	float mx,my;
	mx=( x+.5*sizex)/sizex;
	my=(-y+.5*sizey)/sizey;
	if ((mx>0)&&(mx<1)&&(my>0)&&(my<1)) return 1;
	else return 0;
}

//--------------------------------------------------------------------

int Mesh::collide(Place *place, Physics *physicson) {
	float mx,my, h, xi=0,zi=0, force;
	int hithard=0;
	Place placet;
	Hw *hw=Hw::get();

	mx=( place->pos.x+.5*sizex)/sizex;  // get height @ camera x/z pos --
	my=(-place->pos.z+.5*sizey)/sizey;
	if ((mx<0)||(mx>1)||(my<0)||(my>1)) return 0;
	h=get(mx,my);

	if (place->pos.y<=h) place->physics=physicson; // on mesh? ----------           
	if ((hw->pref.pause==0)&&(hw->pref.help==0)&&(hw->pref.dialogs==0)) place->physics->apply_gravity(place);

	if (place->pos.y<=h) {  // on/under mesh? ---------------------------
		if (fabs(place->vel.y)>hw->pref.gruntMPS) hithard=1;
		place->physics=physicson;
		getv(mx,my, &xi, &zi); // get slide force
		//if (place->physics->ltFriction(xi,0,zi)==0) //{xi=0; zi=0;} // force < minimal friction?
		{
			force=100*hw->pref.meter*hw->time.dt*hw->pref.home*hw->pref.page; // tweak slide force by mesh height --//.01
			place->acc.x+= xi*force; // move camera
			place->acc.z+=-zi*force;
		}
		/*
		if ((place->pos.y-h) < 0) {//-.001*Hw::get()->pref.meter*Hw::get()->time.dt) { // bounce?
			float speed=hw->time.dt, len;

			len=sqrt(xi*xi+100*zi*zi);   // sample slope to offset sphere
			if (len<.00001) len=.00001;
			xi=xi/len;
			zi=zi/len;

			placet.mass=100;                  // tmp hi mass // fake a sphere for collision test
			placet.vel.set_xyz(0,0,0);        // stationary mass 
			placet.pos.x=place->pos.x-xi;     // 
			placet.pos.y=place->pos.y-1;      // just below cam
			placet.pos.z=place->pos.z+zi;     // 
			physicson->bounce(place,&placet); // rotate vel
			place->vel*=1-(2*hw->time.dt);    // damp
			place->pos.y=h;                   // clamp
		}*/

		if ((place->pos.y-h) < -.1*Hw::get()->pref.meter*Hw::get()->time.dt) place->vel.y*=-.3;// below far enough to bounce?
		else                                                                 place->vel.y=0;
		place->pos.y=h;
	}
	return hithard;
}

//---------------------------------------------------------------------

void Mesh::rollup(int type, float f, float radius) {  // apply rollup to mesh func(f)
	int ix,iy, flag;
	float sizexh,sizeyh, fx,fy, dx,dy,dz, cellsizex,cellsizey, f2=f*2, rad;

	sizexh=sizex*.5;
	sizeyh=sizey*.5;
	cellsizex=sizex/nodesx;
	cellsizey=sizey/nodesy;
	for    (ix=0; ix<=nodesx; ix++) {
		for (iy=0; iy<=nodesy; iy++) {
			fx=(float)ix/nodesx;
			fy=(float)iy/nodesy;
			switch(type) {
			case 0: // roll right to left
				if (fx>f) flag=1;
				else      flag=0;
				rad=radius*(1-(fx-f)*.5);
				break;
			case 1: // roll lower right corner up to left
				if ((fx+fy)>f2) flag=1;
				else            flag=0;
				rad=radius*(1-((fx+fy)-f2)*.5);
				break;
			case 2: // roll bottom to top
				if (fy>f) flag=1;
				else      flag=0;
				rad=radius*(1-(fy-f)*.5);
				break;
			}
			if (!flag) {
				dx=ix*cellsizex-sizexh;
				dz=iy*cellsizey-sizeyh;
				h[ix][iy][0]= dx;
				h[ix][iy][1]=  0;
				h[ix][iy][2]=-dz;
			} else {
				switch(type) {
					case 0: // roll right to left
						dx=f*nodesx*cellsizex-sizexh+sinf((fx-f)*pi2*4)*rad;
						dy=                         -cosf((fx-f)*pi2*4)*rad+radius;
						dz=iy*cellsizey-sizeyh;
						break;

					case 1: // roll lower right corner up to left
						//dx=minf(1,((fx-fy)*f*2))*nodesx*cellsizex-sizexh+sinf(((fx+fy)-f2)*pi2*4)*rad;
						//dy=                                             -cosf(((fx+fy)-f2)*pi2*4)*rad+radius;
						//dz=minf(1,((fy-fx)*f*2))*nodesy*cellsizey-sizeyh+sinf(((fx+fy)-f2)*pi2*4)*rad;
						dx=minf(1,((1-fy)*f2))*nodesx*cellsizex-sizexh+sinf(((fx+fy)-f2)*pi2*4)*rad;
						dy=                                           -cosf(((fx+fy)-f2)*pi2*4)*rad+radius;
						dz=minf(1,((1-fx)*f2))*nodesy*cellsizey-sizeyh+sinf(((fx+fy)-f2)*pi2*4)*rad;
						break;

					case 2: // roll bottom to top
						dx=ix*cellsizex-sizexh;
						dy=                         -cosf((fy-f)*pi2*4)*rad+radius;
						dz=f*nodesy*cellsizey-sizeyh+sinf((fy-f)*pi2*4)*rad;
						break;
				}
				h[ix][iy][0]= dx;
				h[ix][iy][1]= dy;
				h[ix][iy][2]=-dz;
			}
		}
	}
}

//-----------------------------------------------------------------------------------

void Mesh::calc_shadow(void) { // 90 deg sun
	static int log=0;
	float v,ht, ambient=Hw::get()->pref.ambient;

	colors=1;
	if (log==0) log_message("   Mesh: Start calc for vertex lighting/shadows...");
	for    (int iy=0; iy<=nodesy; iy++) {
		for (int ix=1; ix<=nodesx; ix++) {
			v=0;
			ht=h[ix][iy][1];
			for (int iz=ix-1; iz>=0; iz--) {
				if ( ht<(h[iz][iy][1]-((ix-iz)*15.)) ) { // 20 short - 10 long
					v=1;
					break;
				}
			}
			if (v==1) {
				i[ix][iy]=ambient;
				//c[ix][iy][0]=.3;
				//c[ix][iy][1]=.3;
				//c[ix][iy][2]=.3;
			}
		}
	}
	if (log==0) {
		log_message("   Mesh: Complete shadow calc.");
		log=1;
	}
}

//------------------------------------------------------------------

int Mesh::cull_strip(int ix, int y0, int y1) {
	Camera *camera=&Hw::get()->screen.viewport->camera;
	if ((camera->cull(h[ix  ][y0][0], h[ix  ][y0][1], h[ix  ][y0][2],10)==-1)&&
	    (camera->cull(h[ix+1][y0][0], h[ix+1][y0][1], h[ix+1][y0][2],10)==-1)&&
	    (camera->cull(h[ix  ][y1][0], h[ix  ][y1][1], h[ix  ][y1][2],10)==-1)&&
	    (camera->cull(h[ix+1][y1][0], h[ix+1][y1][1], h[ix+1][y1][2],10)==-1)
	 )  return 1;
	else return 0;
}

//------------------------------------------------------------------

void Mesh::do_strip(int ix, int y, int y2,float r,float g,float b) {
	int iy;

 	Hw::get()->pref.ntris+=(y2-y+1)*2;
	glBegin(GL_TRIANGLE_STRIP);
		for (iy=y; iy<=y2; iy++) {
			float inten0=i[ix  ][iy],
					inten1=i[ix+1][iy];

			if (normals!=0) glNormal3fv(&n[ix  ][iy][0]);
			if (colors !=0) glColor3f(inten0*r,inten0*g,inten0*b);
			glTexCoord4fv(&t[ix  ][iy][0]); glVertex3fv(&h[ix  ][iy][0]);

			if (normals!=0) glNormal3fv(&n[ix+1][iy][0]); 
			if (colors !=0) glColor3f(inten1*r,inten1*g,inten1*b);
			glTexCoord4fv(&t[ix+1][iy][0]); glVertex3fv(&h[ix+1][iy][0]);
		}
		glEnd();
}

//--------------------------------------------------------------------------


